// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "sweepobject.h"
#include "frontline.h"

#ifndef _GENERATOR_INCLUDE
#define _GENERATOR_INCLUDE

namespace vorosweep
{

class ConeMesh
{
private:
  std::vector<double*> normals;
  std::vector<double*> dirs;
  std::vector<double> speeds;
  std::vector<double*> edges;
  int nf;
public:
  ConeMesh ( int nface )
  {
    nf = nface;
    normals.resize ( nf );
    dirs.resize ( nf );
    speeds.resize ( nf );
    edges.resize ( nf );
    for ( int i = 0; i < nf; i++ )
    {
      normals[i] = new double[3];
      dirs[i] = new double[3];
      edges[i] = new double[3];
    }
  }
  ~ConeMesh ( )
  {
    for ( int i = 0; i < nf; i++ )
    {
      delete normals[i];
      delete dirs[i];
      delete edges[i];
    }
  }
  inline void set_face ( int i, double *n, double *dir )
  {
    geom::copyvec3d ( n, normals[i] );
    geom::copyvec3d ( dir, dirs[i] );
    double z = - ( normals[i][0]*dirs[i][0] + normals[i][1]*dirs[i][1] ) / normals[i][2];
    double xyz = geom::sqrnorm2d ( dirs[i] ) / ( z*z );
    speeds[i] = sqrt ( xyz );

  }
  inline void set_edge ( int i, double *dir )
  {
    geom::copyvec3d ( dir, edges[i] );
  }
  inline void generate_edges()
  {
    for ( int i = 0; i < nf; i++ )
    {
      geom::cross_product3d ( normals[i], normals[ ( i+1 ) % nf], edges[i] );
      geom::normalize3d ( edges[i] );
      dbgprintf ( 2, "edge %d dir --> %lf %lf %lf\n", i, edges[i][0], edges[i][1], edges[i][2] );
    }
  }
  inline int size()
  {
    return nf;
  }
  inline int get_index ( int i )
  {
    return ( i+nf ) %nf;
  }
  inline double *get_normal ( int i )
  {
    int id = get_index ( i );
    return normals[id];
  }
  inline double *get_dir ( int i )
  {
    int id = get_index ( i );
    return dirs[id];
  }
  inline double get_speed ( int i )
  {
    int id = get_index ( i );
    return speeds[id];
  }
  inline double *get_edge ( int i )
  {
    int id = get_index ( i );
    return edges[id];
  }
};

class Generator : public basic_type
{
public:
  static const char *statusname[];

public:
  enum genstatus {ACTIVE, NOACTIVE};
  enum gentype {POINT, LINE};
protected:
  ConeMesh *cone;
  std::vector<SweepEdge *> se;
  std::vector<SweepFacet *> sf;
  double starttime;
  gentype type;
  genstatus state;
	double order;
  double angle;
  double v[2];
  matrix *A;

public:

  Generator ( double t0, int nface )
  {
    cone = new ConeMesh ( nface );
    starttime = t0;
    se.resize ( nface );
    sf.resize ( nface );
    state = Generator::NOACTIVE;
		order = 2.0;
  }
  ~Generator()
  {
    for ( int i = 0; i < ( int ) se.size(); i++ )
      delete se[i];
    for ( int i = 0; i < ( int ) sf.size(); i++ )
      delete sf[i];
  }
  inline double get_start_time()
  {
    return starttime;
  }
  inline int initial_face_size()
  {
    return cone->size();
  }
  inline int face_size()
  {
    return sf.size();
  }
  inline double *get_normal ( int i )
  {
    return cone->get_normal ( i );
  }
  inline double *get_prev_normal ( int i )
  {
    return cone->get_normal ( i-1 );
  }
  inline double *get_next_normal ( int i )
  {
    return cone->get_normal ( i+1 );
  }
  inline double *get_dir ( int i )
  {
    return cone->get_dir ( i );
  }
  inline double *get_prev_dir ( int i )
  {
    return cone->get_dir ( i-1 );
  }
  inline double *get_next_dir ( int i )
  {
    return cone->get_dir ( i+1 );
  }
  inline double *get_edge ( int i, int j )
  {
    return cone->get_edge ( i-j );
  }
  inline double get_speed ( int i )
  {
    return cone->get_speed ( i );
  }
  inline int get_prev_index ( int i )
  {
    return cone->get_index ( i-1 );
  }
  inline int get_next_index ( int i )
  {
    return cone->get_index ( i+1 );
  }
  inline matrix *get_matrix()
  {
    return A;
  }
  inline SweepEdge *get_sweepedge ( int i )
  {
    if ( i >= ( int ) se.size() )
      return NULL;
    return se[i];
  }
  inline SweepFacet *get_sweepfacet ( int i )
  {
    if ( i >= ( int ) sf.size() )
      return NULL;
    return sf[i];
  }
  inline void add_sweepfacet ( SweepFacet *f )
  {
    sf.push_back ( f );
  }
  inline Generator::gentype get_type()
  {
    return type;
  }
  inline int is_active()
  {
    return ( state == ACTIVE );
  }
  inline void set_active()
  {
    state = ACTIVE;
  }
  void update_current ( double time );

  void compute_normal_plane ( double p, double t, double *plane );
  void compute_normal_plane ( double p, double *vec, double *plane );
  void generate_radial_subdivision ( );
  int find_interval ( double *plane, int *index, double *newdir0, double *newdir1 );

};

class Convex_generator : public Generator
{
protected:
  Point *apex;
public:
//   Convex_generator ( double *pt, double t0, int nface, double a, double v1, double v2, int index );
  Convex_generator ( double *pt, double t0, int nface, double a, double v1, double v2, int index ) : Generator ( t0, nface )
  {
    apex = basic_type::pcontainer->get();
//     if ( Parameters::instance()->vorosweep.perturb )
//     {
//       pt[0] += 1e-8*drand48();
//       pt[1] += 1e-8*drand48();
//     }
    apex->set ( pt );
    int id = basic_type::pcontainer->get_index();
    apex->set_index ( id );
    apex->p[2] = t0;
    angle = a;
    v[0] = v1;
    v[1] = v2;
    type = Generator::POINT;
  }
  inline Point *get_apex ( )
  {
    return apex;
  }
  void generate();
};


class Linear_generator : public Generator
{
protected:
  std::vector<Point *> apex;
public:
  Linear_generator ( std::vector<double *> pt, double t0, int nface, double a, double vr, int index ) : Generator ( t0, nface )
  {
    apex.resize ( ( int ) pt.size() );
    for ( int i = 0; i < ( int ) pt.size(); i++ )
    {
      apex[i] = basic_type::pcontainer->get();
      apex[i]->set ( pt[i] );
      int id = basic_type::pcontainer->get_index();
      apex[i]->set_index ( id );
      apex[i]->p[2] = t0;
    }
    angle = a;
    v[0] = v[1] = vr;
    type = Generator::LINE;
  }
  inline std::vector<Point *> &get_apex ( )
  {
    return apex;
  }
  void generate();
};

};

#endif
