// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "elementary.h"
#include "vorosweep.h"
#include <algorithm>

using namespace vorosweep;

class CompareByAngle
{
public:
  /**
   * Usage: std::sort(begin, end, CompareByAngle(origin))
   * -----------------------------------------------------------------------
   * Constructs a new comparator with the indicated point as the origin.
   */
  explicit CompareByAngle ( const Point *orig );

  bool operator() ( SweepEdge *lhs, SweepEdge *rhs ) const;

private:
  const Point *origin;
};

/* Constructor stores the specified origin. */
CompareByAngle::CompareByAngle ( const Point *orig ) : origin ( orig )
{
  // Handled by initializer list
}

/* Comparison function compares first by angle, then by distance. */
bool CompareByAngle::operator() ( SweepEdge *lhs, SweepEdge *rhs ) const
{
  const double normOne = atan2 ( lhs->get_dir2d() [1],-lhs->get_dir2d() [0] );
  const double normTwo = atan2 ( rhs->get_dir2d() [1],-rhs->get_dir2d() [0] );

  return normOne < normTwo;
}

void Graph::handle_multicrash ( Event *ev )
{
  SweepEdge *se = ev->get_sweepedge ( 0 );
  std::vector<SweepEdge*> extras = ev->get_extras();

  double *pos = ev->get_coord();
  double time = ev->get_time();
	double gid = ev->get_gid();

  // build list of non stopped edges
  std::vector<SweepEdge*> elist;
  for ( int i = 0; i < ( int ) extras.size(); i++ )
  {
    extras[i]->print();
    if ( !extras[i]->is_stopped() )
      elist.push_back ( extras[i] );
  }

  if ( elist.size() < 2 )
    return;

  dbgprintf ( 2, "--> %d edges meeting at %.10lf %.10lf\n", ( int ) elist.size(), pos[0], pos[1] );
//   printf ( "--> %d edges meeting at %.10lf %.10lf\n", ( int ) elist.size(), pos[0], pos[1] );
  // sort points by angle
  std::sort ( elist.begin(), elist.end(), CompareByAngle ( se->get_current() ) );
  for ( int i = 0; i < ( int ) elist.size(); i++ )
  {
    SweepEdge *cse = elist[i];
    dbgprintf ( 2, "%p --> start %lf %lf\n", cse, cse->get_start()->p[0], cse->get_start()->p[1] );
//     printf ( "%p --> start %lf %lf\n", cse, cse->get_start()->p[0], cse->get_start()->p[1] );
  }
//   getchar();

  // merge all endpoints
  SweepEdge *master = elist[0];
  for ( int i = 1; i < ( int ) elist.size(); i++ )
  {
    master->merge_current ( elist[i], time, pos );
  }

  for ( int i = 0; i < ( int ) elist.size(); i++ )
  {
    // clockwise rotation
    SweepEdge *se0 = elist[i];
    SweepEdge *se1 = elist[ ( i+1 ) % ( int ) elist.size()];
    ASSERT_ERROR ( se0->is_stopped() && se1->is_stopped(), "Edges have to be stopped\n" );


    dbgprintf ( 2, "se %p --> start %lf %lf\n", se0, se0->get_start()->p[0], se0->get_start()->p[1] );
    dbgprintf ( 2, "se %p --> start %lf %lf\n", se1, se1->get_start()->p[0], se1->get_start()->p[1] );
    se0->set_stopped();
		se0->set_last_gid(gid);
    se1->set_stopped();
		se1->set_last_gid(gid);
    SweepFacet *msf[2] = {NULL, NULL};

    msf[0] = se0->get_sweepfacet ( 0 );
    msf[1] = se1->get_sweepfacet ( 1 );

//     printf ( "msf0 --> %lf %lf %lf\n", msf[0]->get_plane() [0], msf[0]->get_plane() [1], msf[0]->get_plane() [2] );
//     printf ( "msf1 --> %lf %lf %lf\n", msf[1]->get_plane() [0], msf[1]->get_plane() [1], msf[1]->get_plane() [2] );

    if ( msf[0] && msf[1] )
    {
      if ( msf[0] != msf[1] )
      {
        SweepEdge *existing = already_merged ( se0, se1 );
        if ( existing )
        {
          dbgprintf ( 2, "existing : se %p --> start %lf %lf\n", existing, existing->get_start()->p[0], existing->get_start()->p[1] );
          master->merge_current ( existing, time, pos );
        }
        else
        {
          // compute new directions
          double ndir3d[3];
          new_dir ( msf[0], msf[1], ndir3d );
          new_edge ( se0, se1, ndir3d );
        }
      }
      else if ( msf[0] == msf[1] && se0->get_front ( 0 ) != se1->get_front ( 1 ) )
      {
        FrontLine *mfl = msf[0]->get_frontline();
        Front *frn = mfl->front_merge ( se0->get_front ( 0 ), se1->get_front ( 1 ) );
        // generate the new edgecrash
        SweepEdge *frnse0 = frn->get_sweepedge ( 0 );
        SweepEdge *frnse1 = frn->get_sweepedge ( 1 );
        next_edgecrash ( frnse0, frnse1 );
      }
    }
    else if ( msf[0] && !msf[1] ) // relative to border crashes
    {
      merge_and_go_left ( se0, se1, msf[0] );
    }
    else if ( !msf[0] && msf[1] )
    {
      merge_and_go_right ( se0, se1, msf[1] );
    }
  }
}

