// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include <set>
#include <list>
#include <vector>
#include <stack>
#include <map>
#include "common.h"
#include "elementary.h"
#include "parameters.h"
#include "event.h"
#include <geom.h>

#ifndef _SWEEPOBJECT_INCLUDE
#define _SWEEPOBJECT_INCLUDE

namespace vorosweep
{
class Border;
class SweepObject;
class SweepEdge;
class SweepFacet;
class Front;
class FrontLine;
class Event;
class EventPtrComp;
class Convex_generator;

template<class T> class Container
{
private:
  std::stack<T*> freeobject;
  std::stack<int> freeindex;
  int maxindex;
public:
  Container()
  {
    maxindex = 0;
  }
  ~Container()
  {
    while ( !freeobject.empty() )
    {
      T *ob = freeobject.top();
      delete ob;
      freeobject.pop();
    }
  }
  inline T *get ()
  {
    T *ob = NULL;
    if ( freeobject.empty() )
      ob = new T;
    else
    {
      ob = freeobject.top();
      freeobject.pop();
    }
    return ob;
  }
  inline void del ( T *ob )
  {
    freeobject.push ( ob );
  }
  inline int get_index ()
  {
    int index = -1;
    if ( freeindex.empty() )
    {
      index = maxindex++;
    }
    else
    {
      index = freeindex.top();
      freeindex.pop();
    }
    return index;
  }
  inline void del_index ( int index )
  {
    freeindex.push ( index );
  }
};

class basic_type
{
public:
  static Container<Point> *pcontainer;
};

class SweepObject : public basic_type
{
public:
  static const char *objectname[];
  static const char *statusname[];
  static Parameters *pm;
public:
  enum sweeptype {EDGE, BORDEREDGE, FACET};
  enum sweepstatus {STARTING, RUNNING, STOPPED};
protected:
  Point *start; // start position
  double starttime;
  double currenttime;
  double xyspeed; // speed in the Z plane
  double dir3d[3]; // displacement vector
  double dir2d[3]; // displacement vector
  sweeptype type;
  sweepstatus state;
  std::set<Event *, EventPtrComp> eventqueue;

public:
  virtual ~SweepObject()
  {
  }
  inline void init ( Point *from, double t0 )
  {
    dbgprintf ( 1, "new sweepobject %p --> start at %lf %lf %lf, time %lf\n", this, from->p[0], from->p[1], from->p[2], t0 );
    start = from;
    starttime = currenttime = t0;
    dir2d[0] = dir2d[1] = 0.0;
    dir3d[0] = dir3d[1] = dir3d[2] = 0.0;
    state = SweepObject::STARTING;
    xyspeed = 0.0;
  }
  inline void set_dir ( double *vec )
  {
    dir3d[0] = vec[0];
    dir3d[1] = vec[1];
    dir3d[2] = vec[2];
    dir2d[0] = vec[0];
    dir2d[1] = vec[1];
    dir2d[2] = 0.0;
    geom::normalize2d ( dir2d );
  }
  SweepObject ( Point *from, double t0 )
  {
    init ( from, t0 );
  }
  SweepObject ( Point *from, double *vec, double t0 )
  {
    init ( from, t0 );
    set_dir ( vec );
  }
  inline void set_xyspeed ( double s )
  {
    dbgprintf ( 1, "--> setting xy speed %lf\n", s );
    xyspeed = s;
  }
  inline double get_xyspeed ( )
  {
    return xyspeed;
  }
  inline double get_squared_xyspeed ( )
  {
    return xyspeed*xyspeed;
  }
  inline Point *get_start()
  {
    return start;
  }
  inline double *get_dir3d()
  {
    return dir3d;
  }
  inline double *get_dir2d()
  {
    return dir2d;
  }
  inline sweeptype get_type()
  {
    return type;
  }
  inline void update_current ( double time )
  {
    currenttime = time;
  }
//   virtual double time_to_position ( double *pt ) = 0;
  inline double time_to_position ( double *pt )
  {
    double u[2];
    geom::vec2d ( start->p, pt, u );
    double d = geom::scalar_product2d ( get_dir2d(), u );
    return starttime + d / xyspeed;
  }
  Event *get_earliest_event();
  void disable_futur_event ( SweepObject *s0, SweepObject *s1 );
  void remove_earliest_event ( Event *ev );
  inline void insert_event ( Event *ev )
  {
    print_event_queue();
    dbgprintf ( 2, "%s %p --> adding event %p from %d\n", objectname[get_type() ], this, ev, ( int ) eventqueue.size() );
    eventqueue.insert ( ev );
  }
  void update_event(Event *ev, double time, double dist);
  inline void clear_events()
  {
    dbgprintf ( 3, "%s %p --> clearing events\n", objectname[get_type() ], this );
    eventqueue.clear();
  }
  inline void set_running()
  {
    state = SweepObject::RUNNING;
  }
  inline void set_stopped()
  {
    clear_events();
    dbgprintf ( 2, "%s --> stopping sweepobject %p start %lf %lf dir %lf %lf \n", objectname[get_type() ], this, get_start()->p[0], get_start()->p[1], get_dir2d() [0], get_dir2d() [1] );
    state = SweepObject::STOPPED;
  }
  inline int has_started()
  {
    return ( state == SweepObject::STARTING );
  }
  inline int is_stopped()
  {
    if ( state == SweepObject::STOPPED )
      dbgprintf ( 3, "%s --> sweepobject %p start %lf %lf dir %lf %lf is stopped\n", objectname[get_type() ], this, get_start()->p[0], get_start()->p[1], get_dir2d() [0], get_dir2d() [1] );
    return ( state == SweepObject::STOPPED );
  }
  virtual int self_check ()
  {
    return 0;
  }
  virtual void print();
  void print_event_queue();
};


class SweepFacet : public SweepObject
{
public:
//   enum status {INIT, NOINIT};
  enum origtype {POINT, LINE};
private:
  SweepEdge *iedges[2]; // initial sweepedges
//   Point *current; // current position
  FrontLine *mfl;
  double pl[4];
  Generator *cv;
  int radial_index[2]; // radial position in ConeMesh : the second is for linear facet
  origtype origin;
public:
  SweepFacet ( Point *apex, double *plane, double *vec, double t0 ) : SweepObject ( apex, vec, t0 )
  {
    pl[0] = plane[0];
    pl[1] = plane[1];
    pl[2] = plane[2];
    pl[3] = -plane[0]*apex->p[0]-plane[1]*apex->p[1]-plane[2]*apex->p[2];
    type = SweepObject::FACET;
    iedges[0] = iedges[1] = NULL;
    radial_index[0] = radial_index[1] = -1;
    origin = POINT;
  }
//   inline void set_pos ( int p )
//   {
//     radial_index[0] = p;
//   }
//   inline int get_pos()
//   {
//     return radial_index[0];
//   }
  inline void set_pos ( int p, int i=0 )
  {
    radial_index[i] = p;
  }
  inline int get_pos ( int i=0 )
  {
    return radial_index[i];
  }
  inline origtype get_origin()
  {
    return origin;
  }
// 	inline Point *get_start(int i=0)
//   {
//     return iedges[i]->get_start();
//   }
  inline void set_frontline ( FrontLine *fl )
  {
    dbgprintf ( 2, "--> new frontline %p\n", fl );
    mfl = fl;
  }
  inline FrontLine *get_frontline ( )
  {
    return mfl;
  }
  inline void set_initial_sweepedge ( int i, SweepEdge * se )
  {
    dbgprintf ( 2, "--> setting initial sweepedge %d\n", i );
    iedges[i] = se;
  }
  void finalize();
  inline SweepEdge *get_initial_sweepedge ( int i )
  {
    return iedges[i];
  }
//   SweepEdge *get_original_sweepedge ( int i );
  double *get_angular_bound ( int i );
  inline void set_generator ( Generator *c )
  {
    cv = c;
  }
  inline Generator *get_generator()
  {
    return cv;
  }
//   inline double time_to_position ( double *pt )
//   {
//     double u[2];
//     geom::vec2d ( start->p, pt, u );
//     double d = geom::scalar_product2d ( get_dir2d(), u );
//     return starttime + d / xyspeed;
//   }
  inline void position_at_time ( double time, double *pt )
  {
    double d = ( time-starttime ) *xyspeed;
    pt[0] = start->p[0] + d*dir2d[0];
    pt[1] = start->p[1] + d*dir2d[1];
  }
  inline double *get_plane()
  {
    return pl;
  }
  inline void get_ortho_vector ( int i, double *vec2d )
  {
    if ( i == 1 )
    {
      vec2d[0] = -dir2d[1];
      vec2d[1] = dir2d[0];
    }
    else
    {
      vec2d[0] = dir2d[1];
      vec2d[1] = -dir2d[0];
    }
  }
  inline void get_ortho_vector ( double *vec2d )
  {
    vec2d[0] = dir2d[1];
    vec2d[1] = -dir2d[0];
  }
  inline void dir2d_to_dir3d ( double *dir, double *ndir )
  {
    ndir[0] = dir[0];
    ndir[1] = dir[1];
    ndir[2] = - ( pl[0]*ndir[0]+pl[1]*ndir[1] ) / pl[2];
    geom::normalize3d ( ndir );
  }
  inline double compute_squared_speeds ( double *dir )
  {
    double z = - ( get_plane() [0]*dir[0] + get_plane() [1]*dir[1] ) / get_plane() [2];
    double xys = geom::sqrnorm2d ( dir ) / ( z*z );
    return xys;
  }
  int self_check ();
  void print();
};

class SweepEdge : public SweepObject
{
public:
  enum status {INIT, NOINIT};
private:
  Point *current;
  SweepFacet *ifaces[2]; // initial sweepfaces
  Front *fr[2];
  int lgid;
  double stime;
  status init;
  Event *nextswitch;
public:
  SweepEdge ( Point *from, double t0 ) : SweepObject ( from, t0 )
  {
    int id = basic_type::pcontainer->get_index();
    current = pcontainer->get();
    current->set ( start->p );
    current->set_index ( id );
    current->inc_ref();
    type = SweepObject::EDGE;
    fr[0] = fr[1] = NULL;
    ifaces[0] = ifaces[1] = NULL;
    init = SweepEdge::NOINIT;
    lgid = -1;
    stime = 0.0;
    nextswitch = NULL;
  }
  SweepEdge ( Point *from, double *vec, double t0 ) : SweepObject ( from, vec, t0 )
  {
    int id = basic_type::pcontainer->get_index();
    current = pcontainer->get();
    current->set ( start->p );
    current->set_index ( id );
    current->inc_ref();
    type = SweepObject::EDGE;
    fr[0] = fr[1] = NULL;
    ifaces[0] = ifaces[1] = NULL;
    init = SweepEdge::NOINIT;
    lgid = -1;
    stime = 0.0;
  }
  inline Point *get_current()
  {
    return current;
  }
  inline void update_current ( double time, double *pos = NULL )
  {
    SweepObject::update_current ( time );
		if ( pos )
    {
      current->p[0] = pos[0];
      current->p[1] = pos[1];
    }
    else if ( xyspeed != INFINITY )
    {
      current->p[0] = start->p[0] + ( time-starttime ) *xyspeed*dir2d[0];
      current->p[1] = start->p[1] + ( time-starttime ) *xyspeed*dir2d[1];
    }
    current->p[2] = time;
  }
  inline void merge_current ( SweepEdge *se )
  {
    set_stopped();
    se->set_stopped();
    if ( current != se->current )
    {
// 			printf("ref --> %d %d\n", se->get_current()->get_ref(), get_current()->get_ref());
      if ( se->get_current()->get_ref() < 2 )
      {
        basic_type::pcontainer->del ( se->current );
        basic_type::pcontainer->del_index ( se->current->get_index() );
        se->current->set_index ( -1 );
        se->current = current;
      }
      else if ( get_current()->get_ref() < 2 )
      {
        basic_type::pcontainer->del ( current );
        basic_type::pcontainer->del_index ( current->get_index() );
        current->set_index ( -1 );
        current = se->current;
      }
      current->inc_ref();
    }

    dbgprintf ( 2, "--> merging complete : current %d = %p, current %d = %p\n", current->get_index(), current, se->current->get_index(), se->current );
		dbgprintf ( 2, "--> current : %lf %lf\n", current->p[0], current->p[1] );
		ASSERT_ERROR ( current == se->current, "merging vertex failed\n" );
  }
  inline void merge_current ( SweepEdge *se, double time, double *pos = NULL )
  {
    update_current ( time, pos );
    se->currenttime = time;
    merge_current ( se );
  }
  inline void set_last_gid ( int gid )
  {
    lgid = gid;
  }
  inline int get_last_gid()
  {
    return lgid;
  }
  inline Front *get_front ( int i )
  {
    return fr[i];
  }
//   inline void set_next_switch ( double switchtime )
//   {
//     stime = switchtime;
//   }
//   inline double get_next_switch()
//   {
//     return stime;
//   }
  inline void set_next_switch ( Event *ev )
  {
    nextswitch = ev;
  }
  inline double get_next_switch()
  {
    return nextswitch->get_time();
  }
  inline Event *get_next_switch_event()
  {
    return nextswitch;
  }
  inline void set_init()
  {
    init = INIT;
  }
  inline int is_init()
  {
    return ( init == SweepEdge::INIT );
  }
  inline Generator *get_generator()
  {
    if ( is_init() )
      return ifaces[0]->get_generator();
    else
      return NULL;
  }
  inline Generator *get_generator ( int i )
  {
    if ( ifaces[i] )
      return ifaces[i]->get_generator();
    else
      return NULL;
  }
//   inline double time_to_position ( double *pt )
//   {
// //         printf("time to current --> %lf\n", starttime + geom::distance2d ( start->p, current->p ) / xyspeed);
// //     double d = geom::distance2d ( start->p, pt );
//     double u[2];
//     geom::vec2d ( start->p, pt, u );
//     double d = geom::scalar_product2d ( get_dir2d(), u );
//     return starttime + d / xyspeed;
//   }
  inline void position_at_time ( double time, double *pt )
  {
    double d = ( time-starttime ) *xyspeed;
    pt[0] = start->p[0] + d*dir2d[0];
    pt[1] = start->p[1] + d*dir2d[1];
  }
//   inline void position_at_time ( double time, double *pt )
//   {
//     if ( xyspeed == INFINITY )
//     {
//       pt[0] = start->p[0] + dir2d[0];
//       pt[1] = start->p[1] + dir2d[1];
//     }
//     else
//     {
//       double d = ( time-starttime ) *xyspeed;
//       pt[0] = start->p[0] + d*dir2d[0];
//       pt[1] = start->p[1] + d*dir2d[1];
//     }
//   }
//   void remove_earliest_event ( Event *ev );
//   Event *get_earliest_event();
  int compatible_event ( SweepFacet *sf );
  SweepFacet *get_neighbor_sweepfacet ( int i, int level );
  void set_front ( int i, Front *f );
  SweepFacet *get_sweepfacet ( int i );
  void update_speeds ( SweepFacet *sf );
  void update_speeds ( double *plane );
  int self_check ();
  void print();
};

class BorderSweepEdge : public SweepEdge
{
private:
  Border *bo;
public:
  BorderSweepEdge ( Point *from, double t0 ) : SweepEdge ( from, t0 )
  {
    type = SweepObject::BORDEREDGE;
  }
  BorderSweepEdge ( Point *from, double *vec, double t0 ) : SweepEdge ( from, vec, t0 )
  {
    type = SweepObject::BORDEREDGE;
  }
  inline void set_border ( Border *b )
  {
    bo = b;
  }
  inline Border *get_border()
  {
    return bo;
  }
};

};

#endif

