// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "common.h"
#include "parameters.h"
#include "sweepobject.h"
#include "generator.h"
#include "grid.h"
#include "statistics.h"
#include "border.h"
#include <geom.h>
#include "domain2d.h"

#ifndef _GRAPH_INCLUDE
#define _GRAPH_INCLUDE

namespace vorosweep
{


class Graph
{
private:
  Statistic *stats;
  Domain *dom;
  Grid2d *bucketgrid;
  std::set<Event *, EventPtrComp> eventqueue;
  std::set<Event *, EventPtrComp> priorityeventqueue;
  std::vector<Generator *> generatorlist;
  std::vector<Border *> borderlist;
  Container<Event> *econtainer;
  int gnf; // number of face of generators;
  double ctime;

  std::vector<Point *> pointlist;
  std::map<Point *, int> indexmap;
  std::map<Point *, int>::iterator itindex;
  int nbg;
  int nbf;
  int nbvf;
  char checksum[100]; // used to uniquely identify the result

public:
  Graph ( )
  {

    dom = NULL;
    bucketgrid = new Grid2d ( Parameters::instance()->grid.enlarge );
    econtainer = new Container<Event>;
    ctime = 0.0;
    gnf = 0;

    stats = new Statistic;
    srand48 ( time ( NULL ) );

//     exactinit();
  }
  ~Graph()
  {
    for ( int i = 0; i < ( int ) generatorlist.size(); i++ )
      delete generatorlist[i];
    for ( int i = 0; i < ( int ) borderlist.size(); i++ )
      delete borderlist[i];
  }
  inline void add_generator ( Generator *g )
  {
    generatorlist.push_back ( g );
    BBox *bb = bucketgrid->get_bbox();
    if ( g->get_type() == Generator::POINT )
    {
      Convex_generator *cg = reinterpret_cast<Convex_generator *> ( g );
      Point *apex = cg->get_apex();
      dbgprintf ( 2, "apex --> %lf %lf %lf\n", apex->p[0], apex->p[1], apex->p[2] );
      bb->update ( apex->p );
    }
    else if ( g->get_type() == Generator::LINE )
    {
      Linear_generator *lg = reinterpret_cast<Linear_generator *> ( g );
      std::vector<Point *> apex = lg->get_apex();
      for ( int i = 0; i < ( int ) apex.size(); i++ )
        bb->update ( apex[i]->p );
    }
    gnf += g->face_size();
  }
  inline void add_border ( Border *bo )
  {
    borderlist.push_back ( bo );
    BBox *bb = bucketgrid->get_bbox();
    for ( int i = 0; i < 2; i++ )
    {
      Point *pt = bo->get_point ( i );
      dbgprintf ( 2, "pt --> %lf %lf %lf\n", pt->p[0], pt->p[1], pt->p[2] );
      bb->update ( pt->p );
    }
  }
    inline void debug_event_queue ( int nb )
  {
		std::set<Event *, EventPtrComp>::iterator it = eventqueue.begin();;
		for (int i = 0; i < nb; i++)
		{
			Event *ev = *it;
			ev->print();
			it++;
			if (it == eventqueue.end())
				return;
		}
  }
  inline void add_event ( Event *ev )
  {
    if ( ev->get_time() >= ctime ) // future events only
    {
      dbgprintf ( 1, "adding : " );
      ev->print();
			dbgprintf ( 1, "event queue before : %d\n", ( int ) eventqueue.size() );
      ev->it = eventqueue.insert ( ev ).first;
      dbgprintf ( 1, "event queue after : %d\n", ( int ) eventqueue.size() );
// 			debug_event_queue ( 40 );
    }
  }

  inline void update_event ( Event *ev, double time, double dist )
  {
    dbgprintf ( 1, "updating : " );
    ev->print();
//     eventqueue.erase ( ev->it );
		eventqueue.erase ( ev );
// 		debug_event_queue ( 40 );
		ev->update(time, dist);
    add_event ( ev );
  }
  inline void add_priority_event ( Event *ev )
  {
// 		dbgprintf ( 1, "add to priorityeventqueue %.10lf <= %.10lf\n", ev->get_time(), ctime);
    if ( ev->get_time() >= ctime ) // future events only
    {
      dbgprintf ( 1, "adding new item : " );
      ev->print();
      priorityeventqueue.insert ( ev );
    }
  }
  inline int is_queue_empty()
  {
    return ( eventqueue.empty() && priorityeventqueue.empty() );
  }
  inline Event *get_next_event()
  {
    std::set<Event *, EventPtrComp>::iterator it;
    Event *ev = NULL;
    if ( priorityeventqueue.empty() )
    {
      it = eventqueue.begin();
      ev = ( *it );
      eventqueue.erase ( it );
    }
    else
    {
      it = priorityeventqueue.begin();
      ev = ( *it );
      priorityeventqueue.erase ( it );
    }
    return ev;
  }
  inline void update_time ( double time )
  {
    if ( ctime < time )
      ctime = time;
  }
  inline double get_time()
  {
    return ctime;
  }
  inline int get_pointlist_size()
  {
    return pointlist.size();
  }
//     inline unsigned long get_checksum()
//     {
//       return checksum;
//     }
  inline char *get_checksum()
  {
    return checksum;
  }

  void generate();
  void read_input_file ( const char *name );
  void read_borders ( const char *name );
  void default_borders();
  void default_bbox();
  void init();
  void run();
  void fakerun();
  void finalize();
  void export_vtk ( const char *name, bool flat = 0 );
  void export_vector_graphic ( const char *name );
  void export_asymptote ( const char *name );


private:
  unsigned long hash ( const char *str );
  void update_all();
	void update_queue(Event *ev);
  void generate_internals();
  void insert_generator ( Generator *g );
  void add_linear_obstacle ( double *p0, double *p1 );
  void insert_border ( Border *bo );
  void handle ( Event *ev );
  void handle_newgenerator ( Event *ev );
  void handle_newfacetswitch ( Event *ev );
  void handle_newedgeswitch ( Event *ev );
  void handle_borderedgeswitch ( Event *ev );
  void handle_newborderedgeswitch ( Event *ev );
  void handle_edgeswitch ( Event *ev );
  void handle_simple_edgeswitch ( Event *ev );
  void handle_facetswitch ( Event *ev );
  void handle_edge_edgecrash1 ( Event *ev );
  void handle_edge_edgecrash2 ( Event *ev );
  void handle_multicrash ( Event *ev );
  void handle_edge_facetcrash ( Event *ev );
  void handle_edge_bordercrash ( Event *ev );
  void handle_facet_bordercrash ( Event *ev );
  void process_edge_events ( SweepEdge *se, double ctime, int gid );
  void process_facet_events ( SweepFacet *sf, double ctime, int gid );
  void merge_and_go_left ( SweepEdge *se0, SweepEdge *se1, SweepFacet *left );
  void merge_and_go_right ( SweepEdge *se0, SweepEdge *se1, SweepFacet *right );
  void voronoi_vertex_merge ( SweepEdge *se0, SweepEdge *se1, double *ndir );
	void new_edge(SweepEdge *se0, SweepEdge *se1, double *ndir3d);
  void new_facet ( SweepEdge *se0, SweepEdge *se1, double *ndir );
  int enclose_island ( SweepEdge *se0, SweepEdge *se1 );
	void insert_edge_in_bucket(SweepEdge *se, double *pos, double time, int gid);
	void insert_facet_in_bucket(SweepFacet *sf, double *pos, double time, int gid);
  void simple_newedgeswitch ( Point *current, SweepEdge *se0, SweepEdge *se1, double *ndir3d, SweepFacet *msf[2] );
  void merge ( SweepEdge *se0, SweepEdge *se1 );
	void new_dir ( double *n0, double *n1, double *ndir );
  void new_dir ( SweepFacet *sf0, SweepFacet *sf1, double *ndir );
  SweepEdge * already_merged ( SweepEdge *se0, SweepEdge *se1 );
	SweepEdge * already_merged ( Front *fr0, Front *fr1 );
	int check_T_configuration ( SweepEdge *se );
  Event *next_edgeswitch ( SweepEdge *se, double *pos, int gid );
  Event *next_borderedgeswitch ( SweepEdge *se, Border *bo, double *pos, int gid );
  void next_edgecrashes ( SweepEdge *se );
	void next_edgecrash ( SweepEdge *se0, SweepEdge *se1 );
  void next_facetswitch ( SweepFacet *sf, double *pos, int gid, Event **eventlist, int &nbe );
  void add_earliest_event ( SweepObject *so );
  int check_for_corner ( SweepEdge *se, Border *bo, double time );
  void switch_border ( SweepEdge *se, Border *bo, double time );
  void singular_point_rotation_ccw ( SweepEdge *se, Border *nextborder, double time );
  void singular_point_rotation_cw ( SweepEdge *se, Border *nextborder, double time );
  SweepEdge *singular_point_rotation_ccw ( SweepFacet *sf, Border *bo, Point *from, double time );
  SweepEdge *singular_point_rotation_cw ( SweepFacet *sf, Border *bo, Point *from, double time );
  void rotation_ccw ( SweepEdge *se0, SweepEdge *se1 );
  void rotation_cw ( SweepEdge *se0, SweepEdge *se1 );
};

};

#endif
